# Copyright (C) 2014 Okami, okami@fuzetsu.info

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import logging
import radicale.config
import radicale.log

try:
    from configparser import RawConfigParser as ConfigParser
except ImportError:
    from ConfigParser import RawConfigParser as ConfigParser

try:
    from io import StringIO as StringIO
except ImportError:
    from StringIO import StringIO as StringIO

from django.conf import settings


class HashableConfigParser(ConfigParser):
    def __hash__(self):
        output = StringIO()
        self.write(output)
        hash_ = hash(output.getvalue())
        output.close()
        return hash_

# make the config module hashable for django
radicale.config.__class__ = HashableConfigParser


for section, values in settings.DJRADICALE_CONFIG.items():
    for key, value in values.items():
        if not radicale.config.has_section(section):
            radicale.config.add_section(section)
        radicale.config.set(section, key, value)

radicale.log.LOGGER = logging.getLogger('djradicale')

default_app_config = 'djradicale.config.DjRadicaleConfig'



# Copyright (C) 2014 Okami, okami@fuzetsu.info

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import json
import os
import logging

from datetime import datetime

from contextlib import contextmanager
from radicale import ical

from ..models import DBCollection, DBItem, DBProperties

logger = logging.getLogger('djradicale')

ICAL_TYPES = (
    ical.Event,
    ical.Todo,
    ical.Journal,
    ical.Card,
    ical.Timezone,
)


class Collection(ical.Collection):
    @property
    def headers(self):
        return (
            ical.Header("PRODID:-//Radicale//NONSGML Radicale Server//EN"),
            ical.Header("VERSION:%s" % self.version))

    def delete(self):
        DBItem.objects.filter(collection__path=self.path).delete()
        DBCollection.objects.filter(path=self.path).delete()
        DBProperties.objects.filter(path=self.path).delete()

    def append(self, name, text):
        collection, ccreated = DBCollection.objects.get_or_create(
            path=self.path, parent_path=os.path.dirname(self.path))
        item, icreated = DBItem.objects.get_or_create(
            collection=collection, name=name)
        item.text = text
        item.save()

    def remove(self, name):
        DBItem.objects.filter(collection__path=self.path, name=name).delete()

    def replace(self, name, text):
        collection = DBCollection.objects.get(
            path=self.path, parent_path=os.path.dirname(self.path))
        item = DBItem.objects.get(
            collection=collection, name=name)
        item.text = text
        item.save()

    @property
    def text(self):
        return ical.serialize(self.tag, self.headers, self.items.values())

    @classmethod
    def children(cls, path):
        children = list(
            DBCollection.objects
            .filter(parent_path=path or '')
            .values_list('path', flat=True))
        return map(cls, children)

    @classmethod
    def is_node(cls, path):
        result = True
        if path:
            result = (
                DBCollection.objects
                .filter(parent_path=path or '').exists())
        return result

    @classmethod
    def is_leaf(cls, path):
        result = False
        if path:
            result = (
                DBItem.objects
                .filter(collection__path=path or '').exists())
        return result

    @property
    def last_modified(self):
        try:
            collection = DBCollection.objects.get(path=self.path)
        except DBCollection.DoesNotExist:
            pass
        else:
            if collection.last_modified:
                return datetime.strftime(
                    collection.last_modified, '%a, %d %b %Y %H:%M:%S %z')

    @property
    def tag(self):
        with self.props as props:
            if 'tag' not in props:
                if self.path.endswith(('.vcf', '/carddav')):
                    props['tag'] = 'VADDRESSBOOK'
                else:
                    props['tag'] = 'VCALENDAR'
            return props['tag']

    @property
    @contextmanager
    def props(self):
        # On enter
        properties = {}
        try:
            props = DBProperties.objects.get(path=self.path)
        except DBProperties.DoesNotExist:
            pass
        else:
            properties.update(json.loads(props.text))
        old_properties = properties.copy()
        yield properties
        # On exit
        if old_properties != properties and DBCollection.objects.filter(path=self.path).exists():
            props, created = DBProperties.objects.get_or_create(path=self.path)
            props.text = json.dumps(properties)
            props.save()

    @property
    def items(self):
        items = {}
        for item in DBItem.objects.filter(collection__path=self.path):
            items.update(self._parse(item.text, ICAL_TYPES))
        return items



# Copyright (C) 2014 Okami, okami@fuzetsu.info

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import base64
import copy

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.core.urlresolvers import reverse
from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import RedirectView, View
from django.utils.decorators import method_decorator

from radicale import Application


class ApplicationResponse(HttpResponse):
    def start_response(self, status, headers):
        self.status_code = int(status.split(' ')[0])
        for k, v in dict(headers).items():
            self[k] = v


class DjRadicaleView(Application, View):
    http_method_names = [
        'delete',
        'get',
        'head',
        'mkcalendar',
        'mkcol',
        'move',
        'options',
        'propfind',
        'proppatch',
        'put',
        'report',
    ]

    def __init__(self, **kwargs):
        super(DjRadicaleView, self).__init__()
        super(View, self).__init__(**kwargs)

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        if not request.method.lower() in self.http_method_names:
            return self.http_method_not_allowed(request, *args, **kwargs)
        response = ApplicationResponse()
        answer = self(request.META, response.start_response)
        if answer:
            response.write(answer[0])
        return response


class WellKnownView(DjRadicaleView):
    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        # do not authentificate yet, just get the username
        if 'HTTP_AUTHORIZATION' in self.request.META:
            auth = request.META['HTTP_AUTHORIZATION'].split()
            if len(auth) == 2:
                if auth[0].lower() == 'basic':
                    user, password = base64.b64decode(
                        auth[1]).decode().split(':')
                    if kwargs.get('type') == 'carddav':
                        url = '%s/addressbook.vcf/' % user
                    else:
                        url = '%s/calendar.ics/' % user
                    request.META['PATH_INFO'] = reverse(
                        'djradicale:application', kwargs={'url': url})
        return super().dispatch(request, *args, **kwargs)



# Copyright (C) 2014 Okami, okami@fuzetsu.info

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import re

# Manage Python2/3 different modules
# pylint: disable=F0401
try:
    from configparser import RawConfigParser as ConfigParser
except ImportError:
    from ConfigParser import RawConfigParser as ConfigParser
# pylint: enable=F0401

from django.conf import settings

from radicale import log


# Default configuration
INITIAL_RIGHTS = {
    'rw': {
        'user': '.+',
        'collection': '.*',
        'permission': 'rw',
    },
}


def _read_from_sections(user, collection_url, permission):
    regex = ConfigParser({'login': user, 'path': collection_url})
    for rights in (INITIAL_RIGHTS, settings.DJRADICALE_RIGHTS):
        for section, values in rights.items():
            if not regex.has_section(section):
                regex.add_section(section)
            for key, value in values.items():
                regex.set(
                    section, key,
                    value % {
                        'login': re.escape(user),
                        'path': re.escape(collection_url),
                    })
    log.LOGGER.debug("Rights type '%s'" % __name__)

    for section in regex.sections():
        re_user = regex.get(section, 'user')
        re_collection = regex.get(section, 'collection')
        log.LOGGER.debug(
            "Test if '%s:%s' matches against '%s:%s' from section '%s'" % (
                user, collection_url, re_user, re_collection, section))
        user_match = re.match(re_user, user)
        if user_match:
            re_collection = re_collection.format(*user_match.groups())
            if re.match(re_collection, collection_url):
                log.LOGGER.debug("Section '%s' matches" % section)
                if permission in regex.get(section, 'permission'):
                    return True
            else:
                log.LOGGER.debug("Section '%s' does not match" % section)
    return False


def authorized(user, collection, permission):
    collection_url = collection.url.rstrip('/') or '/'
    if collection_url in ('.well-known/carddav', '.well-known/caldav'):
        return permission == 'r'
    return _read_from_sections(user or '', collection_url, permission)



# Copyright (C) 2014 Okami, okami@fuzetsu.info

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

import base64

from django.contrib.auth.models import User
from django.core.urlresolvers import reverse
from django.test import TestCase
from django.test.client import Client


class DAVClient(Client):
    http_authorization = None

    def http_auth(self, username, password):
        auth = '%s:%s' % (username, password)
        self.http_authorization = 'Basic %s' % (
            base64.b64encode(auth.encode())).decode()

    def dispatch(self, method, path, data='', content_type='text/xml',
                 follow=False, secure=False, **extra):
        meta = extra
        if self.http_authorization:
            meta['HTTP_AUTHORIZATION'] = self.http_authorization
        response = self.generic(
            method, path, data=data, content_type=content_type,
            secure=secure, **meta)
        if follow:
            response = self._handle_redirects(response, **meta)
        return response

    def propfind(self, path, data='', content_type='text/xml',
                 follow=False, secure=False, **extra):
        return self.dispatch('PROPFIND', path,
                             data=data, content_type=content_type,
                             follow=follow, secure=secure, **extra)

    def report(self, path, data='', content_type='text/xml',
               follow=False, secure=False, **extra):
        return self.dispatch('REPORT', path,
                             data=data, content_type=content_type,
                             follow=follow, secure=secure, **extra)

    def put(self, path, data='', content_type='text/vcard',
            follow=False, secure=False, **extra):
        return self.dispatch('PUT', path,
                             data=data, content_type=content_type,
                             follow=follow, secure=secure, **extra)

    def get(self, path, data='', content_type='text/xml',
            follow=False, secure=False, **extra):
        return self.dispatch('GET', path,
                             data=data, content_type=content_type,
                             follow=follow, secure=secure, **extra)

    def delete(self, path, data='', content_type='text/xml',
               follow=False, secure=False, **extra):
        return self.dispatch('DELETE', path,
                             data=data, content_type=content_type,
                             follow=follow, secure=secure, **extra)


class DjRadicaleTestCase(TestCase):
    DATA_CASES = {
        # PROPFIND_1 ##########################################################
        'PROPFIND_1': {
            'request': '''<?xml version="1.0"?>
<D:propfind xmlns:D="DAV:" xmlns:x0="http://calendarserver.org/ns/">
    <D:prop>
        <D:resourcetype/>
        <D:supported-report-set/>
        <x0:getctag/>
    </D:prop>
</D:propfind>
''',
            'response': '''<?xml version="1.0"?>
<multistatus xmlns="DAV:" xmlns:CR="urn:ietf:params:xml:ns:carddav" xmlns:CS="http://calendarserver.org/ns/">
  <response>
    <href>{path}</href>
    <propstat>
      <prop>
        <resourcetype>
          <CR:addressbook />
          <collection />
        </resourcetype>
        <supported-report-set>
          <supported-report>
            <report>principal-property-search</report>
          </supported-report>
          <supported-report>
            <report>sync-collection</report>
          </supported-report>
          <supported-report>
            <report>expand-property</report>
          </supported-report>
          <supported-report>
            <report>principal-search-property-set</report>
          </supported-report>
        </supported-report-set>
        <CS:getctag>"d41d8cd98f00b204e9800998ecf8427e"</CS:getctag>
      </prop>
      <status>HTTP/1.1 200 OK</status>
    </propstat>
  </response>
</multistatus>
''',
        },
        # PROPFIND_2 ##########################################################
        'PROPFIND_2': {
            'request': '''<?xml version="1.0" encoding="utf-8" ?>
<D:propfind xmlns:D="DAV:">
    <D:prop>
        <D:displayname/>
        <D:getetag/>
        <D:getcontenttype/>
    </D:prop>
</D:propfind>
''',
            'response': '''<?xml version="1.0"?>
<multistatus xmlns="DAV:">
  <response>
    <href>{path}</href>
    <propstat>
      <prop>
        <displayname>addressbook.vcf</displayname>
        <getetag>"d41d8cd98f00b204e9800998ecf8427e"</getetag>
        <getcontenttype>text/vcard</getcontenttype>
      </prop>
      <status>HTTP/1.1 200 OK</status>
    </propstat>
  </response>
</multistatus>
''',
        },
        # PUT #################################################################
        'PUT': {
            'request': '''BEGIN:VCARD
VERSION:3.0
PRODID:-//Inverse inc.//SOGo Connector 1.0//EN
UID:test.vcf
N:a;a
FN:{fn}
TEL;TYPE=cell:+1234567890
X-MOZILLA-HTML:FALSE
END:VCARD
''',
            'response': '',
        },
        # GET #################################################################
        'GET': {
            'request': '',
            'response': '''BEGIN:VCARD
VERSION:3.0
PRODID:-//Inverse inc.//SOGo Connector 1.0//EN
UID:test.vcf
N:a;a
FN:{fn}
TEL;TYPE=cell:+1234567890
X-MOZILLA-HTML:FALSE
X-RADICALE-NAME:test.vcf
END:VCARD''',
        },
        # REPORT_EMPTY ########################################################
        'REPORT_EMPTY': {
            'request': '''<?xml version="1.0" encoding="utf-8" ?>
<C:addressbook-multiget xmlns:C="urn:ietf:params:xml:ns:carddav">
    <D:prop xmlns:D="DAV:">
        <D:displayname/>
        <D:getetag/>
        <C:address-data/>
    </D:prop>
    <D:href xmlns:D="DAV:">{path}</D:href>
</C:addressbook-multiget>
''',
            'response': '''<?xml version="1.0"?>
<multistatus xmlns="DAV:" />''',
        },
        # REPORT ##############################################################
        'REPORT': {
            'request': '''<?xml version="1.0" encoding="utf-8" ?>
<C:addressbook-multiget xmlns:C="urn:ietf:params:xml:ns:carddav">
    <D:prop xmlns:D="DAV:">
        <D:displayname/>
        <D:getetag/>
        <C:address-data/>
    </D:prop>
    <D:href xmlns:D="DAV:">{path}</D:href>
</C:addressbook-multiget>
''',
            'response': '''<?xml version="1.0"?>
<multistatus xmlns="DAV:" xmlns:CR="urn:ietf:params:xml:ns:carddav">
  <response>
    <href>{path}test.vcf</href>
    <propstat>
      <prop>
        <getetag>"{etag}"</getetag>
        <CR:address-data>BEGIN:VCARD
VERSION:3.0
PRODID:-//Inverse inc.//SOGo Connector 1.0//EN
UID:test.vcf
N:a;a
FN:{fn}
TEL;TYPE=cell:+1234567890
X-MOZILLA-HTML:FALSE
X-RADICALE-NAME:test.vcf
END:VCARD</CR:address-data>
      </prop>
      <status>HTTP/1.1 200 OK</status>
    </propstat>
    <propstat>
      <prop>
        <displayname />
      </prop>
      <status>HTTP/1.1 404 Not Found</status>
    </propstat>
  </response>
</multistatus>
''',
        },
        # DELETE ##############################################################
        'DELETE': {
            'request': '',
            'response': '''<?xml version="1.0"?>
<multistatus xmlns="DAV:">
  <response>
    <href>{path}</href>
    <status>HTTP/1.1 200 OK</status>
  </response>
</multistatus>
''',
        },
    }

    maxDiff = None

    def setUp(self):
        User.objects.create_user(username='user', password='password')
        self.client = DAVClient()

    def propfind_1_anonymous(self):
        path = reverse('djradicale:application', kwargs={
            'url': 'user/addressbook.vcf/',
        })
        response = self.client.propfind(
            path, data=self.DATA_CASES['PROPFIND_1']['request'].format(**{
                'path': path,
            }))
        self.assertEqual(response.status_code, 401)

    def propfind_1(self):
        path = reverse('djradicale:application', kwargs={
            'url': 'user/addressbook.vcf/',
        })
        response = self.client.propfind(
            path, data=self.DATA_CASES['PROPFIND_1']['request'].format(**{
                'path': path,
            }))
        self.assertEqual(response.status_code, 207)
        self.assertEqual(
            response.content.decode(),
            self.DATA_CASES['PROPFIND_1']['response'].format(**{
                'path': path,
            }))

    def propfind_2(self):
        path = reverse('djradicale:application', kwargs={
            'url': 'user/addressbook.vcf/',
        })
        response = self.client.propfind(
            path, data=self.DATA_CASES['PROPFIND_2']['request'].format(**{
                'path': path,
            }))
        self.assertEqual(response.status_code, 207)
        self.assertEqual(
            response.content.decode(),
            self.DATA_CASES['PROPFIND_2']['response'].format(**{
                'path': path,
            }))

    def report_empty(self):
        path = reverse('djradicale:application', kwargs={
            'url': 'user/addressbook.vcf/',
        })
        response = self.client.report(
            path, data=self.DATA_CASES['REPORT_EMPTY']['request'].format(**{
                'path': path,
            }))
        self.assertEqual(response.status_code, 207)
        self.assertEqual(
            response.content.decode(),
            self.DATA_CASES['REPORT_EMPTY']['response'].format(**{
                'path': path,
            }))

    def report(self, fn, etag):
        path = reverse('djradicale:application', kwargs={
            'url': 'user/addressbook.vcf/',
        })
        response = self.client.report(
            path, data=self.DATA_CASES['REPORT']['request'].format(**{
                'path': path,
            }))
        self.assertEqual(response.status_code, 207)
        self.assertEqual(
            response.content.decode(),
            self.DATA_CASES['REPORT']['response'].format(**{
                'path': path,
                'fn': fn,
                'etag': etag,
            }))

    def put(self, fn):
        path = reverse('djradicale:application', kwargs={
            'url': 'user/addressbook.vcf/test.vcf',
        })
        response = self.client.put(
            path, data=self.DATA_CASES['PUT']['request'].format(**{
                'path': path,
                'fn': fn,
            }))
        self.assertEqual(response.status_code, 201)
        self.assertEqual(
            response.content.decode(),
            self.DATA_CASES['PUT']['response'].format(**{
                'path': path,
                'fn': fn,
            }))

    def get(self, fn):
        path = reverse('djradicale:application', kwargs={
            'url': 'user/addressbook.vcf/test.vcf',
        })
        response = self.client.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.content.decode(),
            self.DATA_CASES['GET']['response'].format(**{
                'path': path,
                'fn': fn,
            }))

    def delete(self):
        path = reverse('djradicale:application', kwargs={
            'url': 'user/addressbook.vcf/test.vcf',
        })
        response = self.client.delete(path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.content.decode(),
            self.DATA_CASES['DELETE']['response'].format(**{
                'path': path,
            }))

    def test_everything(self):
        self.propfind_1_anonymous()
        self.client.http_auth('user', 'password')
        self.propfind_1()
        self.propfind_2()
        self.report_empty()
        # create
        self.put(fn='John Smith')
        self.get(fn='John Smith')
        self.report(fn='John Smith', etag='faefbfea2f89431c973bbfd8002d52c8')
        # update
        self.put(fn='J.S.')
        self.get(fn='J.S.')
        self.report(fn='J.S.', etag='6a3d7c7349b462583dba21fbb6321f6b')
        self.delete()
        self.report_empty()
